local AceOO = AceLibrary("AceOO-2.0")
local L = AceLibrary("AceLocale-2.2"):new("GridCustomLayouts")
L:RegisterTranslations("enUS", function()
	return {
		["Custom Layouts"] = true,
		["Add customed layouts using a simple grammer."] = true,
		["Layout text format error, see above information."] = true,
		["Layout name is already used."] = true,
	}
end)
L:RegisterTranslations("zhCN", function()
	return {
		["Custom Layouts"] = "自定义布局",
		["Add customed layouts using a simple grammer."] = "使用简写的格式增加自定义布局, 详见插件说明.",
		["Layout text format error, see above information."] = "布局字符串格式错误, 请看以上信息.",
		["Layout name is already used."] = "创建失败, 布局已经存在.",
	}
end)
L:RegisterTranslations("zhTW", function()
	return {
		["Custom Layouts"] = "自定義佈局",
		["Add customed layouts using a simple grammer."] = "使用簡寫的格式增加自定義佈局, 詳見插件說明.",
		["Layout text format error, see above information."] = "佈局字串格式錯誤, 請看以上資訊.",
		["Layout name is already used."] = "創建失敗, 布局已經存在.",
	}
end)

--[[
List of the various configuration attributes
======================================================
showRaid = [BOOLEAN] -- true if the header should be shown while in a raid
showParty = [BOOLEAN] -- true if the header should be shown while in a party and not in a raid
showPlayer = [BOOLEAN] -- true if the header should show the player when not in a raid
showSolo = [BOOLEAN] -- true if the header should be shown while not in a group (implies showPlayer)
nameList = [STRING] -- a comma separated list of player names (not used if 'groupFilter' is set)
groupFilter = [1-8, STRING] -- a comma seperated list of raid group numbers and/or uppercase class names and/or uppercase roles
strictFiltering = [BOOLEAN] - if true, then characters must match both a group and a class from the groupFilter list
sortMethod = ["INDEX", "NAME"] -- defines how the group is sorted (Default: "INDEX")
sortDir = ["ASC", "DESC"] -- defines the sort order (Default: "ASC")
groupBy = [nil, "GROUP", "CLASS", "ROLE"] - specifies a "grouping" type to apply before regular sorting (Default: nil)
groupingOrder = [STRING] - specifies the order of the groupings (ie. "1,2,3,4,5,6,7,8")
maxColumns = [NUMBER] - maximum number of columns the header will create (Default: 1)
unitsPerColumn = [NUMBER or nil] - maximum units that will be displayed in a singe column, nil is infinate (Default: nil)
--]]


local Dewdrop = AceLibrary:HasInstance("Dewdrop-2.0") and AceLibrary("Dewdrop-2.0")
local waterfall = AceLibrary:HasInstance("Waterfall-1.0") and AceLibrary("Waterfall-1.0")
local BC = LibStub:GetLibrary("LibBabble-Class-3.0"):GetLookupTable()


GridCustomLayouts = Grid:NewModule("GridCustomLayouts")

GridCustomLayouts.defaultDB = {
	layouts = {
		["EXAMPLE1"] = "ZS,DK,QS,LR；groupBy=CLASS\n\nDD,SM,FS,DZ；groupBy=CLASS\n\nMS,SS；groupBy=CLASS\n\n狂暴梧桐，Thatia,三月十二",
		["EXAMPLE2"] = "WAR,DEATH,PAL;groupBy=CLASS;order=WAR,DEATH,PAL;NAME;DESC;showSolo=false;5/2;strictFiltering=false\n1",
	}
}

GridLayout.options.args["GridCustomLayouts"] = {
	type = "execute",
	name = L["Custom Layouts"],
	desc = L["Add customed layouts using a simple grammer."],
	order = 10,
	func = function()
		--GridLayout:SaveLayout("aa", { {groupFilter = "PALADIN",},{groupFilter = "SHAMAN",},{groupFilter = "MAGE",} });
		GridCustomLayoutsFrame:Show()
		local party_type = GridRoster:GetPartyState()
		local layoutName = GridLayout.db.profile.layouts[party_type]
		if ( GridCustomLayouts.db.profile.layouts and GridCustomLayouts.db.profile.layouts[GridLayout.db.profile.layouts[party_type]] ) then
			GridCustomLayouts_SelectLayout(layoutName)
		else
			GridCustomLayouts_SelectFirstLayout()
		end
		GridCustomLayouts_UpdateFrame()
		if(Dewdrop) then Dewdrop:Close() end
		if waterfall and waterfall:IsOpen("Grid") then waterfall:Close("Grid") end
	end
}

function GridLayout:SaveLayout(layoutName, layout)
	if(self.layoutSettings[layoutName]) then
		self.layoutSettings[layoutName] = layout
	else
		self:AddLayout(layoutName, layout)
	end

	if self.db.profile.layouts.force then
		self.db.profile.layouts.force = layoutName
		DEFAULT_CHAT_FRAME:AddMessage("GridCustomLayouts: FORCE LAYOUT set and loaded.", 1, 1, 0);
		self:ReloadLayout()
	else
		local party_type = GridRoster:GetPartyState()
		if ( self.db.profile.layouts[party_type] == layoutName) then
			DEFAULT_CHAT_FRAME:AddMessage("GridCustomLayouts: Layout loaded.", 1, 1, 0);
			self:ReloadLayout()
		end
	end
end

function GridLayout:RemoveLayoutValidate(party_type, layoutName)
	local options = self.options.args[party_type .. "layout"]
	if options then
		local i
		local valid = self.options.args[party_type .. "layout"].validate
		for i=1, #valid do
			if valid[i] == layoutName then
				table.remove(valid, i)
			end
		end
	end
end

function GridLayout:DeleteLayout(layoutName)
	self.layoutSettings[layoutName] = false
	local party_states = {}

	for _, party_type in ipairs(GridRoster.party_states) do
		self:RemoveLayoutValidate(party_type, layoutName)
	end

	self:RemoveLayoutValidate("force", layoutName)
end

function GridLayout:OnInitialize()
	self.super.OnInitialize(self)
	self.layoutGroups = {}
	self.layoutPetGroups = {}



	if GridCustomLayouts and GridCustomLayouts.db then
		for k, v in pairs( GridCustomLayouts.db.profile.layouts ) do
			local layout = GridCustomLayouts_ConvertLayout(v)
			GridLayout:AddLayout(k, layout)
		end
	end
end

local GROUP_FILTER_ABBR = {
	[BC.Warrior] = "WARRIOR",["WARRIOR"] = "WARRIOR",["ZS"] = "WARRIOR",
	[BC.Priest] = "PRIEST",["PRIEST"] = "PRIEST",["MS"] = "PRIEST",
	[BC.Druid] = "DRUID",["DRUID"] = "DRUID",["DD"] = "DRUID",["XD"] = "DRUID",
	[BC.Paladin] = "PALADIN",["PALADIN"] = "PALADIN",["QS"] = "PALADIN",["SQ"] = "PALADIN",
	[BC.Shaman] = "SHAMAN",["SHAMAN"] = "SHAMAN",["SM"] = "SHAMAN",
	[BC.Mage] = "MAGE",["MAGE"] = "MAGE",["FS"] = "MAGE",
	[BC.Warlock] = "WARLOCK",["WARLOCK"] = "WARLOCK",["WL"] = "WARLOCK",["WARLOCK"] = "WARLOCK",["SS"] = "WARLOCK",
	[BC.Hunter] = "HUNTER",	["HUNTER"] = "HUNTER",["LR"] = "HUNTER",
	[BC.Rogue] = "ROGUE",["ROGUE"] = "ROGUE",["DZ"] = "ROGUE",
	[BC.Deathknight] = "DEATHKNIGHT",["DEATHKNIGHT"] = "DEATHKNIGHT",["DK"] = "DEATHKNIGHT",
	["MAINTANK"] = "MAINTANK",["MT"] = "MAINTANK",
	["MAINASSIST"] = "MAINASSIST",["MA"] = "MAINASSIST",
	["1"] = "1",["2"] = "2",["3"] = "3",["4"] = "4",["5"] = "5",["6"] = "6",["7"] = "7",["8"] = "8",
}

function GridCustomLayouts_ConvertOneHeader(line)
	local header = {}
	local attrs = {strsplit(";", line)}
	local haveGroupFilter = false
	for i=1, #attrs do
		local attr = strtrim(attrs[i])
		if(attr=="") then
			--do nothing
		elseif(string.find(attr, "=")) then --attribute=value;
			local t = {strsplit("=", attr)}
			if(#t~=2) then 
				DEFAULT_CHAT_FRAME:AddMessage("Illegal Element: "..attr, 1, 0, 0)
				return nil
			end
			t[1]=strtrim(t[1])
			t[2]=string.gsub(strtrim(t[2]), "\"", "")
			if(strlower(t[1])=="order" or t[1]=="groupingOrder") then
				local t = {strsplit(",", t[2])}
				local j, abbr, av
				for j = 1, #t do
					local v = strtrim(t[j])
					for abbr, av in pairs(GROUP_FILTER_ABBR) do
						if strsub(abbr, 1, #v)==v then --strsub("德鲁伊", 1, #"德")=="德" 
							t[j] = av
							break
						end
					end
				end
				header.groupingOrder = strjoin(",", unpack(t))
			elseif t[1]=="showParty" or t[1]=="showRaid" or t[1]=="showSolo" or t[1]=="showPlayer" or t[1]=="strictFiltering" then
				header[t[1]]= (strlower(t[2]) == "true")  and true or false
			else
				header[t[1]] = t[2]
			end
		elseif(string.find(attr,"/")) then --unitsPerColumn/maxColumns;
			local t = {strsplit("/", attr)}
			if(#t~=2) then 
				DEFAULT_CHAT_FRAME:AddMessage("Illegal Element: "..attr, 1, 0, 0)
				return nil
			end
			header.unitsPerColumn = tonumber(strtrim(t[1]))
			header.maxColumns = tonumber(strtrim(t[2]))
		elseif(attr=="PET") then
			header.isPetGroup = true
		elseif(attr=="INDEX" or attr=="NAME") then
			header.sortMethod = attr
		elseif(attr=="ASC" or attr=="DESC") then
			header.sortDir = attr
		elseif(attr=="STRICT") then
			header.strictFiltering = true
		elseif(attr=="GROUP" or attr=="CLASS" or attr=="ROLE") then
			header.groupBy = attr
		else
			--groupFilter or nameList
			if haveGroupFilter then 
				DEFAULT_CHAT_FRAME:AddMessage("Illegal Element(more than one filter): "..attr, 1,0,0)
				return nil 
			end
			haveGroupFilter = true;
			local t = {strsplit(",", attr)}
			local i, abbr, av, isGroup
			for i = 1, #t do
				local v = strtrim(t[i])
				local thisIsGroup = false
				for abbr, av in pairs(GROUP_FILTER_ABBR) do
					if strsub(abbr, 1, #v)==v then --strsub("德鲁伊", 1, #"德")=="德" 
						thisIsGroup = true
						t[i] = av
						break
					end
				end
				if i==1 then
					isGroup = thisIsGroup
				end

				if isGroup and not thisIsGroup then
					DEFAULT_CHAT_FRAME:AddMessage("Illegal Element(group and name in same filter): "..attr, 1, 0, 0)
					return nil
				end
			end

			if isGroup then
				header.groupFilter = strjoin(",", unpack(t))
				header.groupingOrder = strjoin(",", unpack(t))
			else
				header.nameList = strjoin(",", unpack(t))
			end
		end
	end

	return header
end

--将缩写的格式字符串转换为GridLayoutLayouts的格式
function GridCustomLayouts_ConvertLayout(text)
	text = string.gsub(text,"，", ",")
	text = string.gsub(text,"；", ";")
	text = string.gsub(text,"＝", "=")
	local lines = {strsplit("\n", strtrim(text))}
	local layout = {}
	local count = 1
	for i=1, #lines do
		lines[i] = strtrim(lines[i])
		if strsub(lines[i],1,2)~="--" then
			if(lines[i]=="") then
				if(i>1 and i<#lines and lines[i-1]~="" and lines[i+1]=="") then
					layout[count] = { groupFilter="", } --a double empty line stands for a spacer
					count = count + 1
				end
			else
				layout[count] = GridCustomLayouts_ConvertOneHeader(lines[i])
				if layout[count] == nil then return nil end
				count = count + 1
			end
		end
	end

	return layout
end

function GridCustomLayouts_NewLayout(name)
	if(GridCustomLayouts.db.profile.layouts[name]) then
		DEFAULT_CHAT_FRAME:AddMessage(L["Layout name is already used."], 1, 1, 0)
	else
		GridCustomLayouts.db.profile.layouts[name] = ""
		UIDropDownMenu_Initialize(GridCustomLayoutsFrameDropDown, GridCustomLayouts_DropDown_Initialize)
		GridCustomLayouts_SelectLayout(name)
	end
end

function GridCustomLayouts_SelectLayout(name)
	UIDropDownMenu_Initialize(GridCustomLayoutsFrameDropDown, GridCustomLayouts_DropDown_Initialize)
	UIDropDownMenu_SetSelectedValue(GridCustomLayoutsFrameDropDown, name);
	GridCustomLayoutsFrameArg:SetText(GridCustomLayouts.db.profile.layouts[name] or "")
	GridCustomLayouts_UpdateFrame()
end

function GridCustomLayouts_SelectFirstLayout()
	local name = next(GridCustomLayouts.db.profile.layouts)
	if name then 
		GridCustomLayouts_SelectLayout(name)
	else
		UIDropDownMenu_SetSelectedValue(GridCustomLayoutsFrameDropDown, nil);
	end
	GridCustomLayouts_UpdateFrame()
end
	

function GridCustomLayouts_DeleteLayout()
	local name = UIDropDownMenu_GetSelectedValue(GridCustomLayoutsFrameDropDown)
	if(name) then
		GridCustomLayouts.db.profile.layouts[name] = nil;
		GridLayout:DeleteLayout(name);
		GridCustomLayouts_SelectFirstLayout()
	end
end
		

--=========================================================================
-- FRAME CODE
--=========================================================================
function GridCustomLayoutsFrame_OnLoad(self)
	local frame = self;
	local mover = _G[frame:GetName() .. "Mover"] or CreateFrame("Frame", frame:GetName() .. "Mover", frame)
	mover:EnableMouse(true)
	mover:SetPoint("TOP", frame, "TOP", 0, 10)
	mover:SetWidth(160)
	mover:SetHeight(40)
	mover:SetScript("OnMouseDown", function(self)
		self:GetParent():StartMoving()
	end)
	mover:SetScript("OnMouseUp", function(self)
		self:GetParent():StopMovingOrSizing()
	end)
	frame:SetMovable(true)

	_G[self:GetName().."HeaderText"]:SetText(L["Custom Layouts"])
end

function GridCustomLayoutsFrame_NewOnClick()
	StaticPopup_Show("GRID_NEW_LAYOUT")
end

function GridCustomLayoutsFrame_SaveOnClick()
	local layout = GridCustomLayouts_ConvertLayout(GridCustomLayoutsFrameArg:GetText())
	if layout then
		--DevTools_Dump(GridCustomLayouts_ConvertLayout(GridCustomLayoutsFrameArg:GetText()))
		local layoutName = UIDropDownMenu_GetSelectedValue(GridCustomLayoutsFrameDropDown)
		GridCustomLayouts.db.profile.layouts[layoutName] = GridCustomLayoutsFrameArg:GetText()
		GridLayout:SaveLayout(layoutName, layout)
		GridCustomLayoutsFrameArg:ClearFocus()
	else
		DEFAULT_CHAT_FRAME:AddMessage(L["Layout text format error, see above information."], 1, 0, 0)
	end
end

function GridCustomLayoutsFrame_DeleteOnClick()
	StaticPopup_Show("GRID_LAYOUT_DELETE")
end

function GridCustomLayouts_DropDown_Initialize()
	local info;
	local k,v;
	if GridCustomLayouts and GridCustomLayouts.db then
		for k, _ in pairs( GridCustomLayouts.db.profile.layouts ) do
			info = {};
			info.text = k
			info.func = GridCustomLayouts_DropDown_OnClick;
			info.value = k
			UIDropDownMenu_AddButton(info);
		end
	end
end

function GridCustomLayouts_DropDown_OnClick(self)
	GridCustomLayouts_SelectLayout(self.value)
end

function GridCustomLayouts_UpdateFrame()
	if next(GridCustomLayouts.db.profile.layouts) then
		GridCustomLayoutsFrameSave:Enable()
		GridCustomLayoutsFrameDelete:Enable()
		UIDropDownMenu_EnableDropDown(GridCustomLayoutsFrameDropDown)
	else
		GridCustomLayoutsFrameSave:Disable()
		GridCustomLayoutsFrameDelete:Disable()
		GridCustomLayoutsFrameArg:SetText("")
		UIDropDownMenu_DisableDropDown(GridCustomLayoutsFrameDropDown)
		GridCustomLayoutsFrameDropDownText:SetText("")
	end
	local v = UIDropDownMenu_GetSelectedValue(GridCustomLayoutsFrameDropDown)
	if( not v or GridCustomLayouts.defaultDB.layouts[v]) then
		GridCustomLayoutsFrameDelete:Disable()
	else
		GridCustomLayoutsFrameDelete:Enable()
	end
end

StaticPopupDialogs["GRID_NEW_LAYOUT"] = {
	text = NAME,
	button1 = ACCEPT,
	button2 = CANCEL,
	hasEditBox = 1,
	maxLetters = 24,
	OnAccept = function(self)
		GridCustomLayouts_NewLayout(self.editBox:GetText());
	end,
	OnShow = function(self)
		self.editBox:SetFocus();
	end,
	OnHide = function(self)
		if ( ChatFrameEditBox:IsShown() ) then
			ChatFrameEditBox:SetFocus();
		end
		self.editBox:SetText("");
	end,
	EditBoxOnEnterPressed = function(self)
		local parent = self:GetParent();
		GridCustomLayouts_NewLayout(parent.editBox:GetText());
		parent:Hide();
	end,
	EditBoxOnEscapePressed = function(self)
		self:GetParent():Hide();
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1
};

StaticPopupDialogs["GRID_LAYOUT_DELETE"] = {
	text = CALENDAR_DELETE_EVENT_CONFIRM,
	button1 = ACCEPT,
	button2 = CANCEL,
	OnAccept = function(self)
		GridCustomLayouts_DeleteLayout();
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1
};